//////////////////////////////////////////////////////////////////////
// PSKMod.cpp: implementation of the CPSKMod class.
//
//////////////////////////////////////////////////////////////////////
//      PSK31/CW modulator
//////////////////////////////////////////////////////////////////////
//

#include "stdafx.h"
#include <math.h>
#include "WinPSK.h"
#include "psktables.h"
#include "settings.h"
#include "PSKMod.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// local defines.................

#define PHZ_0 0			//specify various signal phase states
#define PHZ_90 1
#define PHZ_180 2
#define PHZ_270 3
#define PHZ_OFF 4

#define SYM_NOCHANGE 0	//Stay the same phase
#define SYM_P90 1		//Plus 90  deg
#define SYM_P180 2		//Plus 180 deg
#define SYM_M90 3		//Minus 90 deg
#define SYM_OFF 4		//No output
#define SYM_ON 5		//constant output


#define SYMBOL_RATE 31.25		// 31.25 Symbols per Second

#define MAXRAMP_SIZE (((100*11100)/3125)+1) // max number of envelope ramp steps per symbol

#define TX_CONSTANT 23000.0		// TX Amplitude Factor
//#define TX_CONSTANT 10000.0		// TX Amplitude Factor

#define CW_SPEED 2	// wpm = 2.4*dotperiod(in msec) = 2.4*32/(2*CW_SPEED) = 38.4/CW_SPEED

// The use of static's here preclude having multiple instantiations
// of this class but should not be an issue since only one soundcard.

//Ramp shape tables that are loaded with cosine shaped functions at init
static double PSKShapeTbl_Z[MAXRAMP_SIZE];	// 0 
static double PSKShapeTbl_P[MAXRAMP_SIZE];	// +1
static double PSKShapeTbl_M[MAXRAMP_SIZE];	// -1
static double PSKShapeTbl_ZP[MAXRAMP_SIZE];	// 0 to +1
static double PSKShapeTbl_PZ[MAXRAMP_SIZE];	// +1 to 0
static double PSKShapeTbl_MZ[MAXRAMP_SIZE];	// -1 to 0
static double PSKShapeTbl_PM[MAXRAMP_SIZE];	// +1 to -1
static double PSKShapeTbl_MP[MAXRAMP_SIZE];	// -1 to +1

struct PSKStruct
{
	double* iptr;
	double* qptr;
	INT	next;
};

typedef PSKStruct PSKSTATE;

//Lookup table for determining the next ramp shape depending on the
//  next symbol and the present output phase.
// indexing format is [symbol][presentPhase]
//returns the PSKSTRUCT containing the next phase and the I and Q
//   ramp table pointers.
static PSKSTATE PSKPhaseLookupTable[6][5]=
{
// SYMBOL = 0 = SYM_NOCHANGE
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_P, PSKShapeTbl_P, PHZ_0,	//present PHZ_0
	PSKShapeTbl_M, PSKShapeTbl_P, PHZ_90,	//present PHZ_90
	PSKShapeTbl_M, PSKShapeTbl_M, PHZ_180,	//present PHZ_180
	PSKShapeTbl_P, PSKShapeTbl_M, PHZ_270,	//present PHZ_270
	PSKShapeTbl_Z, PSKShapeTbl_Z, PHZ_OFF,	//present PHZ_OFF
// SYMBOL = 1 = SYM_P90 = Advance 90 degrees
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_PM, PSKShapeTbl_P, PHZ_90,	//present PHZ_0
	PSKShapeTbl_M, PSKShapeTbl_PM, PHZ_180,	//present PHZ_90
	PSKShapeTbl_MP, PSKShapeTbl_M, PHZ_270,	//present PHZ_180
	PSKShapeTbl_P, PSKShapeTbl_MP, PHZ_0,	//present PHZ_270
	PSKShapeTbl_ZP, PSKShapeTbl_ZP, PHZ_0,	//present PHZ_OFF
// SYMBOL = 2 = SYM_P180 = Advance 180 degrees
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_PM, PSKShapeTbl_PM, PHZ_180,//present PHZ_0
	PSKShapeTbl_MP, PSKShapeTbl_PM, PHZ_270,//present PHZ_90
	PSKShapeTbl_MP, PSKShapeTbl_MP, PHZ_0,	//present PHZ_180
	PSKShapeTbl_PM, PSKShapeTbl_MP, PHZ_90,	//present PHZ_270
	PSKShapeTbl_ZP, PSKShapeTbl_ZP, PHZ_0,	//present PHZ_OFF
// SYMBOL = 3 = SYM_M90	= retard 90 degrees
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_P, PSKShapeTbl_PM, PHZ_270,	//present PHZ_0
	PSKShapeTbl_MP, PSKShapeTbl_P, PHZ_0,	//present PHZ_90
	PSKShapeTbl_M, PSKShapeTbl_MP, PHZ_90,	//present PHZ_180
	PSKShapeTbl_PM, PSKShapeTbl_M, PHZ_180,	//present PHZ_270
	PSKShapeTbl_ZP, PSKShapeTbl_ZP, PHZ_0,	//present PHZ_OFF
// SYMBOL = 4 = SYM_OFF
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_PZ, PSKShapeTbl_PZ, PHZ_OFF,//present PHZ_0
	PSKShapeTbl_MZ, PSKShapeTbl_PZ, PHZ_OFF,//present PHZ_90
	PSKShapeTbl_MZ, PSKShapeTbl_MZ, PHZ_OFF,//present PHZ_180
	PSKShapeTbl_PZ, PSKShapeTbl_MZ, PHZ_OFF,//present PHZ_270
	PSKShapeTbl_Z, PSKShapeTbl_Z, PHZ_OFF,	//present PHZ_OFF
// SYMBOL = 5 = SYM_ON	
//   I ramp shape     Q ramp shape     Next Phase
	PSKShapeTbl_P, PSKShapeTbl_P, PHZ_0,	//present PHZ_0
	PSKShapeTbl_MP, PSKShapeTbl_P, PHZ_0,	//present PHZ_90
	PSKShapeTbl_MP, PSKShapeTbl_MP, PHZ_0,	//present PHZ_180
	PSKShapeTbl_P, PSKShapeTbl_MP, PHZ_0,	//present PHZ_270
	PSKShapeTbl_ZP, PSKShapeTbl_ZP, PHZ_0	//present PHZ_OFF
};
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPSKMod::CPSKMod()
{
	m_2PI = 8.0 * atan(1.0);		//make constant 2 PI
	m_AmblePtr = 0;
}

CPSKMod::~CPSKMod()
{
}

///////////++++++++++++++++++++++++++++++++++////////////////
///////////   P S K 3 1   M O D U L A T O R  ////////////////
///////////++++++++++++++++++++++++++++++++++////////////////

/////////////////////////////////////////////////////////////
//Initialize the PSK/CW modulator with the sample frequncy Fs,
// the carrier frequency freq, and the PSK mode.
/////////////////////////////////////////////////////////////
void CPSKMod::InitPSKMod(INT Fs, double freq, CWinPSKDoc* pDoc)
{
	m_pDoc = pDoc;
	m_SampleFreq = (double)Fs;
	m_PSKPhaseInc = m_2PI * freq/(double)Fs;		//carrier frequency
	m_PSKmode = m_pDoc->m_pSettings->m_ModType;		//save off local copy of mode
	if( m_pDoc->m_TXState == TX_CWID_STATE )
		m_PSKmode = CW_MODE;
	m_PSKSecPerSamp = 1.0/(double)Fs;
	m_PSKTime = 0.0;
	m_t = 0.0;
	m_Ramp = 0;
	m_PSKPeriodUpdate = 1.0/SYMBOL_RATE;	//symbol period
	m_Lastsymb = SYM_OFF;
	m_AddEndingZero = TRUE;
	m_CWState = 0;
	m_CWtimer = 0;
// Generate cosine ramp envelope lookup tables
	INT RampSize =  (((100*Fs)/3125)+1); //  number of envelope ramp steps per symbol
	for( INT i=0; i<RampSize; i++)
	{
		PSKShapeTbl_Z[i] = 0.0;
		PSKShapeTbl_P[i] = 1.0;
		PSKShapeTbl_M[i] = -1.0;
		PSKShapeTbl_PM[i] = cos( (double)i*m_2PI/(RampSize*2) );
		PSKShapeTbl_MP[i] = -PSKShapeTbl_PM[i];

		if( i <RampSize/2 )
		{
			PSKShapeTbl_PZ[i] = cos( (double)i*m_2PI/(RampSize*2) );
			PSKShapeTbl_MZ[i] = -PSKShapeTbl_PZ[i];
			PSKShapeTbl_ZP[i] = 0.0;
		}
		else
		{
			PSKShapeTbl_ZP[i] = -cos( (double)i*m_2PI/(RampSize*2) );
			PSKShapeTbl_PZ[i] = 0.0;
			PSKShapeTbl_MZ[i] = 0.0;
		}

	}
	i = 0;
	while(i<32)		//create post/preamble tables
	{
		m_Preamble[i] = TXTOG_CODE;
		m_Postamble[i++] = TXON_CODE;
	}
	m_Preamble[i] = 0;		// null terminate these tables
	m_Postamble[i] = 0;
	
	m_pPSKtxI = PSKShapeTbl_Z;
	m_pPSKtxQ = PSKShapeTbl_Z;
	m_PresentPhase = PHZ_OFF;
	m_TxShiftReg = 0;
	m_TxCodeWord = 0;
	m_TestString.LoadString(IDS_TESTMODESTRING);	//load up test string
}


/////////////////////////////////////////////////////////////
// generates n samples of psk31 waveform in data array pData
/////////////////////////////////////////////////////////////
void CPSKMod::CalcPSK(double *pData, INT n)
{
INT symbol;
INT i;
	if( m_PSKmode != CW_MODE)
		m_PSKmode = m_pDoc->m_pSettings->m_ModType;	//change mode on fly if not CW
	if( m_pDoc->m_pSettings->m_DemoMode )
	{
		m_PSKPhaseInc = m_2PI * m_pDoc->m_pSettings->m_TxFreq/m_SampleFreq;	//track TX carrier frequency
		m_RMSConstant = m_pDoc->m_SignalAmplitude;	//get amplitude for noise simulation
	}
	else
		m_RMSConstant = TX_CONSTANT;

	if( (m_pDoc->m_TXState >= TX_SENDING_STATE) &&
					(m_pDoc->m_TXState <= TX_PREAMBLE_STATE) )
	{	// tell Xmit text window to fill up the character Queue
		::PostMessage( m_pDoc->m_hXmitEditCtrl, WM_USER_FILL_QUEUE ,0,0);
	}
	for( i=0; i<n; i++ )		//calculate n samples of tx data stream
	{
		m_t += m_PSKPhaseInc;			// increment radian phase count
// create sample from sin/cos and shape tables
		pData[i] = m_RMSConstant*( m_pPSKtxI[m_Ramp]*sin( m_t ) + m_pPSKtxQ[m_Ramp++]*cos( m_t ) );
		m_PSKTime += m_PSKSecPerSamp;
		if( m_PSKTime >= m_PSKPeriodUpdate )//if time to update envelope ramp index
		{
			m_PSKTime -= m_PSKPeriodUpdate;	//keep time bounded
			m_Ramp = 0;						// time to update symbol
			m_t = fmod(m_t,m_2PI);			//keep radian counter bounded
			switch( m_PSKmode )				//get next symbol to send
			{
				case CW_MODE:
					symbol = GetNextCWSymbol();
					break;
				case BPSK_MODE:
					symbol = GetNextBPSKSymbol();
					break;
				case QPSK_MODE:
					symbol = GetNextQPSKSymbol();
					if(	m_pDoc->m_pSettings->m_UseLSB)
						if(symbol==SYM_P90)		//rotate vectors the opposite way
							symbol = SYM_M90;
						else
							if(symbol==SYM_M90)
							symbol = SYM_P90;
					break;
			}
			//get new I/Q ramp tables and next phase
			m_pPSKtxI = PSKPhaseLookupTable[symbol][m_PresentPhase].iptr;
			m_pPSKtxQ = PSKPhaseLookupTable[symbol][m_PresentPhase].qptr;
			m_PresentPhase = PSKPhaseLookupTable[symbol][m_PresentPhase].next;
		}
	}
}


/////////////////////////////////////////////////////////////
// called every symbol time to get next CW symbol and get the
// next character from the character Queue if no more symbols
// are left to send.
/////////////////////////////////////////////////////////////
BYTE CPSKMod::GetNextCWSymbol(void)
{
BYTE symb;
INT ch;
	symb = m_Lastsymb;		//use last symbol unless it needs to change
	if( (m_TxShiftReg == 0 ) && (m_CWState == 0) )
	{
		ch = GetChar();			//get next character to xmit
		if( ch >=0 )			//if is not a control code
		{
			ch &= 0xFF;		
			ch = (INT)toupper( (BYTE)ch );	//make upper case
			if( ch>=' ' && ch<='Z')
				m_TxShiftReg = CW_TABLE[ ch-' '];	//look up pattern
		}
		else					// is a control code
		{
			if( ch == TXON_CODE )
				symb = SYM_ON;
			else
				symb = SYM_OFF;
			return symb;
		}
		m_CWState = 0;
	}
	switch( m_CWState )		// CW timing state machine
	{
		case 0:			//get next cw symbol state from pattern
			switch( m_TxShiftReg&0xC000 )
			{
				case 0x4000:	//dot
					m_CWState = 1;
					m_CWtimer = 1*CW_SPEED;
					symb = SYM_ON;
					break;
				case 0x8000:	//dash
					m_CWState = 1;
					m_CWtimer = 3*CW_SPEED;
					symb = SYM_ON;
					break;
				case 0xC000:	//inter char space 
					m_CWState = 2;
					m_CWtimer = 2*CW_SPEED-1;
					symb = SYM_OFF;
					break;
				default:
					symb = SYM_OFF;
					break;
			}
			m_TxShiftReg = m_TxShiftReg<<2;	//
			break;
		case 1:		//On time state
			if( --m_CWtimer <= 0 )
			{
				m_CWtimer = 1*CW_SPEED - 1;	//inter symbol time
				if(m_CWtimer>=1)
					m_CWState = 2;
				else
					m_CWState = 0;
				symb = SYM_OFF;
			}
			else
				symb = SYM_NOCHANGE;
			break;
		case 2:		//Off time state
			if( --m_CWtimer <= 0 )
				m_CWState = 0;
			break;
	}
	m_Lastsymb = symb;
	return symb;
}

/////////////////////////////////////////////////////////////
// called every symbol time to get next BPSK symbol and get the
// next character from the character Queue if no more symbols
// are left to send.
/////////////////////////////////////////////////////////////
BYTE CPSKMod::GetNextBPSKSymbol(void)
{
BYTE symb;
INT ch;
	symb = m_Lastsymb;
	if( m_TxShiftReg == 0 )
	{
		if( m_AddEndingZero )		// if is end of code
		{
			symb = SYM_P180;		// end with a zero
			m_AddEndingZero = FALSE;
		}
		else
		{
			ch = GetChar();			//get next character to xmit
			if( ch >=0 )			//if is not a control code
			{						//get next VARICODE codeword to send
				m_TxShiftReg = VARICODE_TABLE[ ch&0xFF ];
				symb = SYM_P180;	//Start with a zero
			}
			else					// is a control code
			{
				switch( ch )
				{
				case TXON_CODE:
					symb = SYM_ON;
					break;
				case TXTOG_CODE:
					symb = SYM_P180;
					break;
				case TXOFF_CODE:
					symb = SYM_OFF;
					break;
				}
			}
		}
	}
	else			// is not end of code word so send next bit
	{
		if( m_TxShiftReg&0x8000 )
			symb = SYM_NOCHANGE;
		else
			symb = SYM_P180;
		m_TxShiftReg = m_TxShiftReg<<1;	//point to next bit
		if( m_TxShiftReg == 0 )			// if at end of codeword
			m_AddEndingZero = TRUE;		// need to send a zero nextime
	}
	m_Lastsymb = symb;
	return symb;
}

		
/////////////////////////////////////////////////////////////
// called every symbol time to get next QPSK symbol and get the
// next character from the character Queue if no more symbols
// are left to send.
/////////////////////////////////////////////////////////////
BYTE CPSKMod::GetNextQPSKSymbol(void)
{
BYTE symb;
INT ch;
	symb = ConvolutionCodeTable[m_TxShiftReg&0x1F];	//get next convolution code
	m_TxShiftReg = m_TxShiftReg<<1;
	if( m_TxCodeWord == 0 )			//need to get next codeword
	{
		if( m_AddEndingZero )		//if need to add a zero
		{
			m_AddEndingZero = FALSE;	//end with a zero
		}
		else
		{
			ch = GetChar();			//get next character to xmit
			if( ch >=0 )			//if not a control code
			{						//get next VARICODE codeword to send
				m_TxCodeWord = VARICODE_TABLE[ ch&0xFF ];
			}
			else					//is a control code
			{
				switch( ch )
				{
				case TXON_CODE:
					symb = SYM_ON;
					break;
				case TXTOG_CODE:
					m_TxCodeWord = 0;
					break;
				case TXOFF_CODE:
					symb = SYM_OFF;
					break;
				}
			}
		}
	}
	else
	{
		if(m_TxCodeWord&0x8000 )
		{
			m_TxShiftReg |= 1;
		}
		m_TxCodeWord = m_TxCodeWord<<1;
		if(m_TxCodeWord == 0)
			m_AddEndingZero = TRUE;	//need to add another zero
	}
	return symb;
}

/////////////////////////////////////////////////////////////
//get next character/symbol depending on TX state. 
/////////////////////////////////////////////////////////////
INT CPSKMod::GetChar()
{
INT ch;
static INT last=0;
	switch( m_pDoc->m_TXState )
	{
		case TX_OFF_STATE:		//is receiving
			if( m_pDoc->m_pSettings->m_DemoMode )	//test mode
				ch = MakeTestChar();	//generate some test characters
			else
				ch = TXOFF_CODE;		//else turn off
			break;
		case TX_TUNE_STATE:
			ch = TXON_CODE;				// steady carrier
			break;
		case TX_PAUSED_STATE:
			ch = TXTOG_CODE;			// steady idle symbol
			break;
		case TX_POSTAMBLE_STATE:		// ending sequence
			if( !(ch = m_Postamble[m_AmblePtr++] ))
			{
				if(	m_pDoc->m_NeedCWid  )
				{
					m_pDoc->m_TXState = TX_CWID_STATE;
					m_PSKmode = CW_MODE;
					m_pDoc->m_NeedCWid = FALSE;
					m_AmblePtr = 0;
					ch = TXOFF_CODE;
				}
				else
				{
					m_pDoc->m_TXState = TX_OFF_STATE;
					m_AmblePtr = 0;
					ch = TXOFF_CODE;
				}
			}
			break;
		case TX_PREAMBLE_STATE:			//starting sequence
			if( !(ch = m_Preamble[m_AmblePtr++] ))
			{
				m_pDoc->m_TXState = TX_SENDING_STATE;
				m_AmblePtr = 0;
				ch = TXTOG_CODE;
			}
			break;
		case TX_CWID_STATE:				// id sendign CW ID
			if( m_AmblePtr >= m_pDoc->m_CWIdString.GetLength() )
			{
				m_pDoc->m_TXState = TX_OFF_STATE;
				m_AmblePtr = 0;
				ch = TXOFF_CODE;
			}
			else
			{
				m_PSKmode = CW_MODE;
				ch = (INT)m_pDoc->m_CWIdString.GetAt(m_AmblePtr++);
			}
			break;
		case TX_SENDING_STATE:		//if sending text from TX window
			ch = m_pDoc->GetTxChar();
			if(	m_pDoc->m_XmitWindowEmpty && ch == TXTOG_CODE)
				m_pDoc->m_XmitFinished = TRUE;
			m_AmblePtr = 0;

			if( (ch > 0 ) && m_pDoc->m_pSettings->m_EchoTXText)
				::PostMessage(m_pDoc->m_hRcvEditCtrl,WM_USER_RCV_CHAR ,ch,XMIT_FONT_ID);
			break;

	}
	last = ch;
	return( ch );
}

/////////////////////////////////////////////////////////////
//  Pulls next character out of teststring then repeats
/////////////////////////////////////////////////////////////
BYTE CPSKMod::MakeTestChar()
{
BYTE ch;
static INT indx = 0;
INT y;
	if( indx >= (y=m_TestString.GetLength()) )
	{
		indx = 0;
		ch = 0x0D;
	}
	else
		ch = m_TestString[indx++];
//ch = TXON_CODE;
//ch = TXTOG_CODE;
//ch = (rand()/345) + ' ';
	return ch;

}
///////////////////////////////////////////////////////////////////////
